@@ -2,10 +2,6 @@
  * Copyright © 2016-2018 The TokTok team.
  * Copyright © 2013-2015 Tox project.
  */
-#ifdef HAVE_CONFIG_H
-#include "config.h"
-#endif /* HAVE_CONFIG_H */
-
 #include "rtp.h"
 
 #include <assert.h>
@@ -13,11 +9,16 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include "bwcontroller.h"
+#include <sodium.h>
 
-#include "../toxcore/Messenger.h"
+#include "bwcontroller.h"
+#include "toxav_hacks.h"
+
+#include "../toxcore/ccompat.h"
 #include "../toxcore/logger.h"
 #include "../toxcore/mono_time.h"
+#include "../toxcore/net_crypto.h"
+#include "../toxcore/tox_private.h"
 #include "../toxcore/util.h"
 
 /**
@@ -26,31 +27,15 @@
  */
 #define VIDEO_KEEP_KEYFRAME_IN_BUFFER_FOR_MS 15
 
-
-/*
- * return -1 on failure, 0 on success
- *
- */
-static int rtp_send_custom_lossy_packet(Tox *tox, int32_t friendnumber, const uint8_t *data, uint32_t length)
-{
-    Tox_Err_Friend_Custom_Packet error;
-    tox_friend_send_lossy_packet(tox, friendnumber, data, (size_t)length, &error);
-
-    if (error == TOX_ERR_FRIEND_CUSTOM_PACKET_OK) {
-        return 0;
-    }
-
-    return -1;
-}
-
 // allocate_len is NOT including header!
-static struct RTPMessage *new_message(const struct RTPHeader *header, size_t allocate_len, const uint8_t *data,
-                                      uint16_t data_length)
+static struct RTPMessage *new_message(const Logger *log, const struct RTPHeader *header, size_t allocate_len,
+                                      const uint8_t *data, uint16_t data_length)
 {
     assert(allocate_len >= data_length);
     struct RTPMessage *msg = (struct RTPMessage *)calloc(1, sizeof(struct RTPMessage) + allocate_len);
 
     if (msg == nullptr) {
+        LOGGER_DEBUG(log, "Could not allocate RTPMessage buffer");
         return nullptr;
     }
 
@@ -196,7 +181,7 @@ static int8_t get_slot(const Logger *log, struct RTPWorkBufferList *wkbl, bool i
  *
  * If there are no frames ready, we return NULL. If this function returns
  * non-NULL, it transfers ownership of the message to the caller, i.e. the
- * caller is responsible for storing it elsewhere or calling free().
+ * caller is responsible for storing it elsewhere or calling `free()`.
  */
 static struct RTPMessage *process_frame(const Logger *log, struct RTPWorkBufferList *wkbl, uint8_t slot_id)
 {
@@ -222,7 +207,7 @@ static struct RTPMessage *process_frame(const Logger *log, struct RTPWorkBufferL
     struct RTPMessage *const m_new = slot->buf;
     slot->buf = nullptr;
 
-    assert(wkbl->next_free_entry >= 1);
+    assert(wkbl->next_free_entry >= 1 && wkbl->next_free_entry <= USED_RTP_WORKBUFFER_COUNT);
 
     if (slot_id != wkbl->next_free_entry - 1) {
         // The slot is not the last slot, so we created a gap. We move all the
@@ -245,7 +230,7 @@ static struct RTPMessage *process_frame(const Logger *log, struct RTPWorkBufferL
 }
 
 /**
- * @param log A logger.
+ * @param log A pointer to the Logger object.
  * @param wkbl The list of in-progress frames, i.e. all the slots.
  * @param slot_id The slot we want to fill the data into.
  * @param is_keyframe Whether the data is part of a key frame.
@@ -263,7 +248,7 @@ static bool fill_data_into_slot(const Logger *log, struct RTPWorkBufferList *wkb
     struct RTPWorkBuffer *const slot = &wkbl->work_buffer[slot_id];
 
     assert(header != nullptr);
-    assert(is_keyframe == (bool)(header->flags & RTP_KEY_FRAME));
+    assert(is_keyframe == (bool)((header->flags & RTP_KEY_FRAME) != 0));
 
     if (slot->received_len == 0) {
         assert(slot->buf == nullptr);
@@ -313,18 +298,18 @@ static bool fill_data_into_slot(const Logger *log, struct RTPWorkBufferList *wkb
     return slot->received_len == header->data_length_full;
 }
 
-static void update_bwc_values(const Logger *log, RTPSession *session, const struct RTPMessage *msg)
+static void update_bwc_values(RTPSession *session, const struct RTPMessage *msg)
 {
     if (session->first_packets_counter < DISMISS_FIRST_LOST_VIDEO_PACKET_COUNT) {
         ++session->first_packets_counter;
     } else {
-        uint32_t data_length_full = msg->header.data_length_full; // without header
-        uint32_t received_length_full = msg->header.received_length_full; // without header
+        const uint32_t data_length_full = msg->header.data_length_full; // without header
+        const uint32_t received_length_full = msg->header.received_length_full; // without header
         bwc_add_recv(session->bwc, data_length_full);
 
         if (received_length_full < data_length_full) {
-            LOGGER_DEBUG(log, "BWC: full length=%u received length=%d", data_length_full, received_length_full);
-            bwc_add_lost(session->bwc, (data_length_full - received_length_full));
+            LOGGER_DEBUG(session->log, "BWC: full length=%u received length=%d", data_length_full, received_length_full);
+            bwc_add_lost(session->bwc, data_length_full - received_length_full);
         }
     }
 }
@@ -348,24 +333,19 @@ static void update_bwc_values(const Logger *log, RTPSession *session, const stru
  *   this is the actual payload length.
  * @param log A logger.
  *
- * @return -1 on error, 0 on success.
+ * @retval -1 on error.
+ * @retval 0 on success.
  */
-static int handle_video_packet(RTPSession *session, const struct RTPHeader *header,
-                               const uint8_t *incoming_data, uint16_t incoming_data_length, const Logger *log)
+static int handle_video_packet(const Logger *log, RTPSession *session, const struct RTPHeader *header,
+                               const uint8_t *incoming_data, uint16_t incoming_data_length)
 {
     // Full frame length in bytes. The frame may be split into multiple packets,
     // but this value is the complete assembled frame size.
     const uint32_t full_frame_length = header->data_length_full;
 
-    // Current offset in the frame. If this is the first packet of a multipart
-    // frame or it's not a multipart frame, then this value is 0.
-    const uint32_t offset = header->offset_full; // without header
-
     // The sender tells us whether this is a key frame.
     const bool is_keyframe = (header->flags & RTP_KEY_FRAME) != 0;
 
-    LOGGER_DEBUG(log, "-- handle_video_packet -- full lens=%u len=%u offset=%u is_keyframe=%s",
-                 (unsigned)incoming_data_length, (unsigned)full_frame_length, (unsigned)offset, is_keyframe ? "K" : ".");
     LOGGER_DEBUG(log, "wkbl->next_free_entry:003=%d", session->work_buffer_list->next_free_entry);
 
     const bool is_multipart = full_frame_length != incoming_data_length;
@@ -390,10 +370,13 @@ static int handle_video_packet(RTPSession *session, const struct RTPHeader *head
         // get_slot just told us it's full, so process_frame must return non-null.
         assert(m_new != nullptr);
 
-        LOGGER_DEBUG(log, "-- handle_video_packet -- CALLBACK-001a b0=%d b1=%d", (int)m_new->data[0], (int)m_new->data[1]);
-        update_bwc_values(log, session, m_new);
+        LOGGER_DEBUG(log, "-- handle_video_packet -- CALLBACK-001a b0=%d b1=%d", (int)m_new->data[0],
+                     (int)m_new->data[1]);
+        update_bwc_values(session, m_new);
         // Pass ownership of m_new to the callback.
-        session->mcb(session->m->mono_time, session->cs, m_new);
+        Mono_Time *mt = toxav_get_av_mono_time(session->toxav);
+        assert(mt != nullptr);
+        session->mcb(mt, session->cs, m_new);
         // Now we no longer own m_new.
         m_new = nullptr;
 
@@ -427,10 +410,13 @@ static int handle_video_packet(RTPSession *session, const struct RTPHeader *head
 
     struct RTPMessage *m_new = process_frame(log, session->work_buffer_list, slot_id);
 
-    if (m_new) {
-        LOGGER_DEBUG(log, "-- handle_video_packet -- CALLBACK-003a b0=%d b1=%d", (int)m_new->data[0], (int)m_new->data[1]);
-        update_bwc_values(log, session, m_new);
-        session->mcb(session->m->mono_time, session->cs, m_new);
+    if (m_new != nullptr) {
+        LOGGER_DEBUG(log, "-- handle_video_packet -- CALLBACK-003a b0=%d b1=%d", (int)m_new->data[0],
+                     (int)m_new->data[1]);
+        update_bwc_values(session, m_new);
+        Mono_Time *mt = toxav_get_av_mono_time(session->toxav);
+        assert(mt != nullptr);
+        session->mcb(mt, session->cs, m_new);
 
         m_new = nullptr;
     }
@@ -439,84 +425,118 @@ static int handle_video_packet(RTPSession *session, const struct RTPHeader *head
 }
 
 /**
- * @return -1 on error, 0 on success.
+ * receive custom lossypackets and process them. they can be incoming audio or video packets
  */
-static int handle_rtp_packet(Messenger *m, uint32_t friendnumber, const uint8_t *data, uint16_t length, void *object)
+void handle_rtp_packet(Tox *tox, uint32_t friend_number, const uint8_t *data, size_t length, void *user_data)
 {
-    RTPSession *session = (RTPSession *)object;
+    ToxAV *toxav = (ToxAV *)tox_get_av_object(tox);
 
-    if (!session || length < RTP_HEADER_SIZE + 1) {
-        LOGGER_WARNING(m->log, "No session or invalid length of received buffer!");
-        return -1;
+    if (toxav == nullptr) {
+        // LOGGER_WARNING(log, "ToxAV is NULL!");
+        return;
+    }
+
+    const Logger *log = toxav_get_logger(toxav);
+
+    if (length < RTP_HEADER_SIZE + 1) {
+        LOGGER_WARNING(log, "Invalid length of received buffer!");
+        return;
+    }
+
+    ToxAVCall *call = call_get(toxav, friend_number);
+
+    if (call == nullptr) {
+        LOGGER_WARNING(log, "ToxAVCall is NULL!");
+        return;
+    }
+
+    RTPSession *session = rtp_session_get(call, data[0]);
+
+    if (session == nullptr) {
+        LOGGER_WARNING(log, "No session!");
+        return;
+    }
+
+    if (!session->rtp_receive_active) {
+        LOGGER_WARNING(log, "receiving not allowed!");
+        return;
     }
 
     // Get the packet type.
     const uint8_t packet_type = data[0];
-    ++data;
-    --length;
+    const uint8_t *payload = &data[1];
+    // TODO(Zoff): is this ok?
+    const uint16_t payload_size = (uint16_t)length - 1;
 
     // Unpack the header.
     struct RTPHeader header;
-    rtp_header_unpack(data, &header);
+    rtp_header_unpack(payload, &header);
 
     if (header.pt != packet_type % 128) {
-        LOGGER_WARNING(m->log, "RTPHeader packet type and Tox protocol packet type did not agree: %d != %d",
+        LOGGER_WARNING(log, "RTPHeader packet type and Tox protocol packet type did not agree: %d != %d",
                        header.pt, packet_type % 128);
-        return -1;
+        return;
     }
 
     if (header.pt != session->payload_type % 128) {
-        LOGGER_WARNING(m->log, "RTPHeader packet type does not match this session's payload type: %d != %d",
+        LOGGER_WARNING(log, "RTPHeader packet type does not match this session's payload type: %d != %d",
                        header.pt, session->payload_type % 128);
-        return -1;
+        return;
     }
 
-    if (header.flags & RTP_LARGE_FRAME && header.offset_full >= header.data_length_full) {
-        LOGGER_ERROR(m->log, "Invalid video packet: frame offset (%u) >= full frame length (%u)",
+    if ((header.flags & RTP_LARGE_FRAME) != 0 && header.offset_full >= header.data_length_full) {
+        LOGGER_ERROR(log, "Invalid video packet: frame offset (%u) >= full frame length (%u)",
                      (unsigned)header.offset_full, (unsigned)header.data_length_full);
-        return -1;
+        return;
     }
 
     if (header.offset_lower >= header.data_length_lower) {
-        LOGGER_ERROR(m->log, "Invalid old protocol video packet: frame offset (%u) >= full frame length (%u)",
+        LOGGER_ERROR(log, "Invalid old protocol video packet: frame offset (%u) >= full frame length (%u)",
                      (unsigned)header.offset_lower, (unsigned)header.data_length_lower);
-        return -1;
+        return;
     }
 
-    LOGGER_DEBUG(m->log, "header.pt %d, video %d", (uint8_t)header.pt, (RTP_TYPE_VIDEO % 128));
+    LOGGER_DEBUG(log, "header.pt %d, video %d", (uint8_t)header.pt, RTP_TYPE_VIDEO % 128);
 
     // The sender uses the new large-frame capable protocol and is sending a
     // video packet.
-    if ((header.flags & RTP_LARGE_FRAME) && header.pt == (RTP_TYPE_VIDEO % 128)) {
-        return handle_video_packet(session, &header, data + RTP_HEADER_SIZE, length - RTP_HEADER_SIZE, m->log);
+    if ((header.flags & RTP_LARGE_FRAME) != 0 && header.pt == (RTP_TYPE_VIDEO % 128)) {
+        handle_video_packet(log, session, &header, &payload[RTP_HEADER_SIZE], payload_size - RTP_HEADER_SIZE);
+        return;
     }
 
     // everything below here is for the old 16 bit protocol ------------------
 
-    if (header.data_length_lower == length - RTP_HEADER_SIZE) {
+    if (header.data_length_lower == payload_size - RTP_HEADER_SIZE) {
         /* The message is sent in single part */
 
         /* Message is not late; pick up the latest parameters */
         session->rsequnum = header.sequnum;
         session->rtimestamp = header.timestamp;
-        bwc_add_recv(session->bwc, length);
+        bwc_add_recv(session->bwc, payload_size);
 
         /* Invoke processing of active multiparted message */
-        if (session->mp) {
-            session->mcb(session->m->mono_time, session->cs, session->mp);
+        if (session->mp != nullptr) {
+            Mono_Time *mt = toxav_get_av_mono_time(session->toxav);
+            assert(mt != nullptr);
+            session->mcb(mt, session->cs, session->mp);
             session->mp = nullptr;
         }
 
         /* The message came in the allowed time;
          */
 
-        return session->mcb(session->m->mono_time, session->cs, new_message(&header, length - RTP_HEADER_SIZE,
-                            data + RTP_HEADER_SIZE, length - RTP_HEADER_SIZE));
+        session->mp = new_message(log, &header, payload_size - RTP_HEADER_SIZE, &payload[RTP_HEADER_SIZE], payload_size - RTP_HEADER_SIZE);
+        Mono_Time *mt = toxav_get_av_mono_time(session->toxav);
+        assert(mt != nullptr);
+        session->mcb(mt, session->cs, session->mp);
+        session->mp = nullptr;
+        return;
     }
 
     /* The message is sent in multiple parts */
 
-    if (session->mp) {
+    if (session->mp != nullptr) {
         /* There are 2 possible situations in this case:
          *      1) being that we got the part of already processing message.
          *      2) being that we got the part of a new/old message.
@@ -529,24 +549,26 @@ static int handle_rtp_packet(Messenger *m, uint32_t friendnumber, const uint8_t
             /* First case */
 
             /* Make sure we have enough allocated memory */
-            if (session->mp->header.data_length_lower - session->mp->len < length - RTP_HEADER_SIZE ||
+            if (session->mp->header.data_length_lower - session->mp->len < payload_size - RTP_HEADER_SIZE ||
                     session->mp->header.data_length_lower <= header.offset_lower) {
                 /* There happened to be some corruption on the stream;
                  * continue wihtout this part
                  */
-                return 0;
+                return;
             }
 
-            memcpy(session->mp->data + header.offset_lower, data + RTP_HEADER_SIZE,
-                   length - RTP_HEADER_SIZE);
-            session->mp->len += length - RTP_HEADER_SIZE;
-            bwc_add_recv(session->bwc, length);
+            memcpy(session->mp->data + header.offset_lower, &payload[RTP_HEADER_SIZE],
+                   payload_size - RTP_HEADER_SIZE);
+            session->mp->len += payload_size - RTP_HEADER_SIZE;
+            bwc_add_recv(session->bwc, payload_size);
 
             if (session->mp->len == session->mp->header.data_length_lower) {
                 /* Received a full message; now push it for the further
                  * processing.
                  */
-                session->mcb(session->m->mono_time, session->cs, session->mp);
+                Mono_Time *mt = toxav_get_av_mono_time(session->toxav);
+                assert(mt != nullptr);
+                session->mcb(mt, session->cs, session->mp);
                 session->mp = nullptr;
             }
         } else {
@@ -555,17 +577,19 @@ static int handle_rtp_packet(Messenger *m, uint32_t friendnumber, const uint8_t
                 /* The received message part is from the old message;
                  * discard it.
                  */
-                return 0;
+                return;
             }
 
             /* Push the previous message for processing */
-            session->mcb(session->m->mono_time, session->cs, session->mp);
+            Mono_Time *mt = toxav_get_av_mono_time(session->toxav);
+            assert(mt != nullptr);
+            session->mcb(mt, session->cs, session->mp);
 
             session->mp = nullptr;
             goto NEW_MULTIPARTED;
         }
     } else {
-        /* In this case threat the message as if it was received in order
+        /* In this case treat the message as if it was received in order
          */
         /* This is also a point for new multiparted messages */
 NEW_MULTIPARTED:
@@ -573,15 +597,21 @@ NEW_MULTIPARTED:
         /* Message is not late; pick up the latest parameters */
         session->rsequnum = header.sequnum;
         session->rtimestamp = header.timestamp;
-        bwc_add_recv(session->bwc, length);
+        bwc_add_recv(session->bwc, payload_size);
 
         /* Store message.
          */
-        session->mp = new_message(&header, header.data_length_lower, data + RTP_HEADER_SIZE, length - RTP_HEADER_SIZE);
-        memmove(session->mp->data + header.offset_lower, session->mp->data, session->mp->len);
+        session->mp = new_message(log, &header, header.data_length_lower, &payload[RTP_HEADER_SIZE], payload_size - RTP_HEADER_SIZE);
+
+        if (session->mp != nullptr) {
+            memmove(session->mp->data + header.offset_lower, session->mp->data, session->mp->len);
+        } else {
+            LOGGER_WARNING(log, "new_message() returned a null pointer");
+            return;
+        }
     }
 
-    return 0;
+    return;
 }
 
 size_t rtp_header_pack(uint8_t *const rdata, const struct RTPHeader *header)
@@ -643,24 +673,29 @@ size_t rtp_header_unpack(const uint8_t *data, struct RTPHeader *header)
     return p - data;
 }
 
-RTPSession *rtp_new(int payload_type, Messenger *m, Tox *tox, uint32_t friendnumber,
+static uint32_t rtp_random_u32(void)
+{
+    // HINT: uses libsodium function
+    return randombytes_random();
+}
+
+RTPSession *rtp_new(const Logger *log, int payload_type, Tox *tox, ToxAV *toxav, uint32_t friendnumber,
                     BWController *bwc, void *cs, rtp_m_cb *mcb)
 {
     assert(mcb != nullptr);
     assert(cs != nullptr);
-    assert(m != nullptr);
 
     RTPSession *session = (RTPSession *)calloc(1, sizeof(RTPSession));
 
-    if (!session) {
-        LOGGER_WARNING(m->log, "Alloc failed! Program might misbehave!");
+    if (session == nullptr) {
+        LOGGER_WARNING(log, "Alloc failed! Program might misbehave!");
         return nullptr;
     }
 
     session->work_buffer_list = (struct RTPWorkBufferList *)calloc(1, sizeof(struct RTPWorkBufferList));
 
     if (session->work_buffer_list == nullptr) {
-        LOGGER_ERROR(m->log, "out of memory while allocating work buffer list");
+        LOGGER_ERROR(log, "out of memory while allocating work buffer list");
         free(session);
         return nullptr;
     }
@@ -668,11 +703,12 @@ RTPSession *rtp_new(int payload_type, Messenger *m, Tox *tox, uint32_t friendnum
     // First entry is free.
     session->work_buffer_list->next_free_entry = 0;
 
-    session->ssrc = payload_type == RTP_TYPE_VIDEO ? 0 : random_u32();
+    session->ssrc = payload_type == RTP_TYPE_VIDEO ? 0 : rtp_random_u32(); // Zoff: what is this??
     session->payload_type = payload_type;
-    session->m = m;
     session->tox = tox;
+    session->toxav = toxav;
     session->friend_number = friendnumber;
+    session->rtp_receive_active = true;
 
     // set NULL just in case
     session->mp = nullptr;
@@ -683,136 +719,146 @@ RTPSession *rtp_new(int payload_type, Messenger *m, Tox *tox, uint32_t friendnum
     session->cs = cs;
     session->mcb = mcb;
 
-    if (-1 == rtp_allow_receiving(session)) {
-        LOGGER_WARNING(m->log, "Failed to start rtp receiving mode");
-        free(session->work_buffer_list);
-        free(session);
-        return nullptr;
-    }
-
     return session;
 }
 
-void rtp_kill(RTPSession *session)
+void rtp_kill(const Logger *log, RTPSession *session)
 {
-    if (!session) {
+    if (session == nullptr) {
+        LOGGER_WARNING(log, "No session");
         return;
     }
 
-    LOGGER_DEBUG(session->m->log, "Terminated RTP session: %p", (void *)session);
-    rtp_stop_receiving(session);
-
-    LOGGER_DEBUG(session->m->log, "Terminated RTP session V3 work_buffer_list->next_free_entry: %d",
+    LOGGER_DEBUG(log, "Terminated RTP session: %p", (void *)session);
+    LOGGER_DEBUG(log, "Terminated RTP session V3 work_buffer_list->next_free_entry: %d",
                  (int)session->work_buffer_list->next_free_entry);
 
+    for (int8_t i = 0; i < session->work_buffer_list->next_free_entry; ++i) {
+        free(session->work_buffer_list->work_buffer[i].buf);
+    }
     free(session->work_buffer_list);
     free(session);
 }
 
-int rtp_allow_receiving(RTPSession *session)
+void rtp_allow_receiving_mark(RTPSession *session)
 {
-    if (session == nullptr) {
-        return -1;
+    if (session != nullptr) {
+        session->rtp_receive_active = true;
     }
-
-    if (m_callback_rtp_packet(session->m, session->friend_number, session->payload_type,
-                              handle_rtp_packet, session) == -1) {
-        LOGGER_WARNING(session->m->log, "Failed to register rtp receive handler");
-        return -1;
-    }
-
-    LOGGER_DEBUG(session->m->log, "Started receiving on session: %p", (void *)session);
-    return 0;
 }
 
-int rtp_stop_receiving(RTPSession *session)
+void rtp_stop_receiving_mark(RTPSession *session)
 {
-    if (session == nullptr) {
-        return -1;
+    if (session != nullptr) {
+        session->rtp_receive_active = false;
     }
-
-    m_callback_rtp_packet(session->m, session->friend_number, session->payload_type, nullptr, nullptr);
-
-    LOGGER_DEBUG(session->m->log, "Stopped receiving on session: %p", (void *)session);
-    return 0;
 }
 
-/**
- * @param data is raw vpx data.
- * @param length is the length of the raw data.
- */
-int rtp_send_data(RTPSession *session, const uint8_t *data, uint32_t length,
-                  bool is_keyframe, const Logger *log)
+void rtp_allow_receiving(Tox *tox)
 {
-    if (!session) {
-        LOGGER_ERROR(log, "No session!");
-        return -1;
-    }
-
-    struct RTPHeader header = {0};
-
-    header.ve = 2;  // this is unused in toxav
-
-    header.pe = 0;
-
-    header.xe = 0;
-
-    header.cc = 0;
-
-    header.ma = 0;
-
-    header.pt = session->payload_type % 128;
-
-    header.sequnum = session->sequnum;
-
-    header.timestamp = current_time_monotonic(session->m->mono_time);
-
-    header.ssrc = session->ssrc;
-
-    header.offset_lower = 0;
-
-    // here the highest bits gets stripped anyway, no need to do keyframe bit magic here!
-    header.data_length_lower = length;
-
-    if (session->payload_type == RTP_TYPE_VIDEO) {
-        header.flags = RTP_LARGE_FRAME;
+    // register callback
+    tox_callback_friend_lossy_packet_per_pktid(tox, handle_rtp_packet, RTP_TYPE_AUDIO);
+    tox_callback_friend_lossy_packet_per_pktid(tox, handle_rtp_packet, RTP_TYPE_VIDEO);
+}
+
+void rtp_stop_receiving(Tox *tox)
+{
+    // UN-register callback
+    tox_callback_friend_lossy_packet_per_pktid(tox, nullptr, RTP_TYPE_AUDIO);
+    tox_callback_friend_lossy_packet_per_pktid(tox, nullptr, RTP_TYPE_VIDEO);
+}
+
+static void rtp_send_piece(const Logger *log, Tox *tox, uint32_t friend_number, const struct RTPHeader *header,
+                           const uint8_t *data, uint8_t *rdata, uint16_t length)
+{
+    rtp_header_pack(rdata + 1, header);
+    memcpy(rdata + 1 + RTP_HEADER_SIZE, data, length);
+
+    Tox_Err_Friend_Custom_Packet error;
+    tox_friend_send_lossy_packet(tox, friend_number,
+                                 rdata, length + RTP_HEADER_SIZE + 1, &error);
+
+    if (error != TOX_ERR_FRIEND_CUSTOM_PACKET_OK) {
+        char *netstrerror = net_new_strerror(net_error());
+        LOGGER_WARNING(log, "RTP send failed (len: %d)! tox error: %d, net error: %s",
+                       length + RTP_HEADER_SIZE + 1, error, netstrerror);
+        net_kill_strerror(netstrerror);
     }
+}
 
+static struct RTPHeader rtp_default_header(const RTPSession *session, uint32_t length, bool is_keyframe)
+{
     uint16_t length_safe = (uint16_t)length;
 
     if (length > UINT16_MAX) {
         length_safe = UINT16_MAX;
     }
 
-    header.data_length_lower = length_safe;
-    header.data_length_full = length; // without header
-    header.offset_lower = 0;
-    header.offset_full = 0;
+    struct RTPHeader header = {0};
 
     if (is_keyframe) {
         header.flags |= RTP_KEY_FRAME;
     }
 
-    VLA(uint8_t, rdata, length + RTP_HEADER_SIZE + 1);
-    memset(rdata, 0, SIZEOF_VLA(rdata));
+    if (session->payload_type == RTP_TYPE_VIDEO) {
+        header.flags |= RTP_LARGE_FRAME;
+    }
+
+    header.ve = 2;  // this is unused in toxav
+    header.pe = 0;
+    header.xe = 0;
+    header.cc = 0;
+    header.ma = 0;
+    header.pt = session->payload_type % 128;
+    header.sequnum = session->sequnum;
+    Mono_Time *mt = toxav_get_av_mono_time(session->toxav);
+    if (mt != nullptr) {
+        header.timestamp = current_time_monotonic(mt);
+    } else {
+        header.timestamp = 0;
+    }
+    header.ssrc = session->ssrc;
+    header.offset_lower = 0;
+    header.data_length_lower = length_safe;
+    header.data_length_full = length; // without header
+    header.offset_lower = 0;
+    header.offset_full = 0;
+
+    return header;
+}
+
+/**
+ * @brief Send a frame of audio or video data, chunked in @ref RTPMessage instances.
+ *
+ * @param session The A/V session to send the data for.
+ * @param data A byte array of length @p length.
+ * @param length The number of bytes to send from @p data.
+ * @param is_keyframe Whether this video frame is a key frame. If it is an
+ *   audio frame, this parameter is ignored.
+ */
+int rtp_send_data(const Logger *log, RTPSession *session, const uint8_t *data, uint32_t length,
+                  bool is_keyframe)
+{
+    if (session == nullptr) {
+        return -1;
+    }
+
+    const uint16_t rdata_size = min_u32(length + RTP_HEADER_SIZE + 1, MAX_CRYPTO_DATA_SIZE);
+    VLA(uint8_t, rdata, rdata_size);
+    memset(rdata, 0, rdata_size);
     rdata[0] = session->payload_type;  // packet id == payload_type
 
+    struct RTPHeader header = rtp_default_header(session, length, is_keyframe);
+
     if (MAX_CRYPTO_DATA_SIZE > (length + RTP_HEADER_SIZE + 1)) {
-        /**
+        /*
          * The length is lesser than the maximum allowed length (including header)
          * Send the packet in single piece.
          */
-        rtp_header_pack(rdata + 1, &header);
-        memcpy(rdata + 1 + RTP_HEADER_SIZE, data, length);
-
-        if (-1 == rtp_send_custom_lossy_packet(session->tox, session->friend_number, rdata, SIZEOF_VLA(rdata))) {
-            const char *netstrerror = net_new_strerror(net_error());
-            LOGGER_WARNING(session->m->log, "RTP send failed (len: %u)! std error: %s, net error: %s",
-                           (unsigned)SIZEOF_VLA(rdata), strerror(errno), netstrerror);
-            net_kill_strerror(netstrerror);
-        }
+        assert(length < UINT16_MAX);
+        rtp_send_piece(log, session->tox, session->friend_number, &header, data, rdata, length);
     } else {
-        /**
+        /*
          * The length is greater than the maximum allowed length (including header)
          * Send the packet in multiple pieces.
          */
@@ -820,16 +866,7 @@ int rtp_send_data(RTPSession *session, const uint8_t *data, uint32_t length,
         uint16_t piece = MAX_CRYPTO_DATA_SIZE - (RTP_HEADER_SIZE + 1);
 
         while ((length - sent) + RTP_HEADER_SIZE + 1 > MAX_CRYPTO_DATA_SIZE) {
-            rtp_header_pack(rdata + 1, &header);
-            memcpy(rdata + 1 + RTP_HEADER_SIZE, data + sent, piece);
-
-            if (-1 == rtp_send_custom_lossy_packet(session->tox, session->friend_number,
-                                                   rdata, piece + RTP_HEADER_SIZE + 1)) {
-                const char *netstrerror = net_new_strerror(net_error());
-                LOGGER_WARNING(session->m->log, "RTP send failed (len: %d)! std error: %s, net error: %s",
-                               piece + RTP_HEADER_SIZE + 1, strerror(errno), netstrerror);
-                net_kill_strerror(netstrerror);
-            }
+            rtp_send_piece(log, session->tox, session->friend_number, &header, data + sent, rdata, piece);
 
             sent += piece;
             header.offset_lower = sent;
@@ -839,17 +876,8 @@ int rtp_send_data(RTPSession *session, const uint8_t *data, uint32_t length,
         /* Send remaining */
         piece = length - sent;
 
-        if (piece) {
-            rtp_header_pack(rdata + 1, &header);
-            memcpy(rdata + 1 + RTP_HEADER_SIZE, data + sent, piece);
-
-            if (-1 == rtp_send_custom_lossy_packet(session->tox, session->friend_number, rdata,
-                                                   piece + RTP_HEADER_SIZE + 1)) {
-                const char *netstrerror = net_new_strerror(net_error());
-                LOGGER_WARNING(session->m->log, "RTP send failed (len: %d)! std error: %s, net error: %s",
-                               piece + RTP_HEADER_SIZE + 1, strerror(errno), netstrerror);
-                net_kill_strerror(netstrerror);
-            }
+        if (piece != 0) {
+            rtp_send_piece(log, session->tox, session->friend_number, &header, data + sent, rdata, piece);
         }
     }
 
